@@ -1,7 +1,5 @@
 import datetime
 import functools
-import hashlib
-import inspect
 import sys
 import time
 import uuid
@@ -11,24 +9,41 @@ import platform
 import warnings
 import types
 import numbers
+import inspect
 
 from dateutil import parser
 from dateutil.tz import tzlocal
 
+
+try:
+    from maya import MayaDT
+except ImportError:
+    MayaDT = None
+
+
 real_time = time.time
 real_localtime = time.localtime
 real_gmtime = time.gmtime
 real_strftime = time.strftime
+real_clock = time.clock
 real_date = datetime.date
 real_datetime = datetime.datetime
 real_date_objects = [real_time, real_localtime, real_gmtime, real_strftime, real_date, real_datetime]
 _real_time_object_ids = set(id(obj) for obj in real_date_objects)
 
+times_to_freeze = []
+tz_offsets = []
 
 try:
     real_uuid_generate_time = uuid._uuid_generate_time
-except (AttributeError, ImportError):
+    uuid_generate_time_attr = '_uuid_generate_time'
+except AttributeError:
+    uuid._load_system_functions()
+    real_uuid_generate_time = uuid._generate_time_safe
+    uuid_generate_time_attr = '_generate_time_safe'
+except ImportError:
     real_uuid_generate_time = None
+    uuid_generate_time_attr = None
 
 try:
     real_uuid_create = uuid._UuidCreate
@@ -51,18 +66,7 @@ except AttributeError:
 
 
 # keep a cache of module attributes otherwise freezegun will need to analyze too many modules all the time
-# start with `None` as the sentinel value.
-# if `{}` (empty dict) was the sentinel value, there's a chance that `setup_modules_cache()` will be called many times
-_GLOBAL_MODULES_CACHE = None
-
-
-def _get_global_modules_cache():
-    global _GLOBAL_MODULES_CACHE
-    # the first call to this function sets up the global module cache. it's expected to be slower than consecutive calls
-    if _GLOBAL_MODULES_CACHE is None:
-        _GLOBAL_MODULES_CACHE = {}
-        _setup_modules_cache()
-    return _GLOBAL_MODULES_CACHE
+_GLOBAL_MODULES_CACHE = {}
 
 
 def _get_module_attributes(module):
@@ -82,16 +86,7 @@ def _get_module_attributes(module):
     return result
 
 
-def _setup_modules_cache():
-    for mod_name, module in list(sys.modules.items()):
-        # ignore modules from freezegun
-        if mod_name == __name__ or not mod_name or not module or not hasattr(module, "__name__"):
-            continue
-        _setup_module_cache(module)
-
-
 def _setup_module_cache(module):
-    global _GLOBAL_MODULES_CACHE
     date_attrs = []
     all_module_attributes = _get_module_attributes(module)
     for attribute_name, attribute_value in all_module_attributes:
@@ -101,17 +96,23 @@ def _setup_module_cache(module):
 
 
 def _get_module_attributes_hash(module):
-    return '{0}-{1}'.format(id(module), hashlib.md5(','.join(dir(module)).encode('utf-8')).hexdigest())
+    try:
+        module_dir = dir(module)
+    except TypeError:
+        module_dir = []
+    return '{0}-{1}'.format(id(module), hash(frozenset(module_dir)))
 
 
-def _get_cached_module_attributes(mod_name, module):
-    global_modules_cache = _get_global_modules_cache()
-    module_hash, cached_attrs = global_modules_cache.get(mod_name, ('0', []))
+def _get_cached_module_attributes(module):
+    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE.get(module.__name__, ('0', []))
     if _get_module_attributes_hash(module) == module_hash:
         return cached_attrs
-    else:
-        _setup_module_cache(module)
-        return _get_module_attributes(module)
+
+    # cache miss: update the cache and return the refreshed value
+    _setup_module_cache(module)
+    # return the newly cached value
+    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE[module.__name__]
+    return cached_attrs
 
 
 # Stolen from six
@@ -125,51 +126,108 @@ _is_cpython = (
 )
 
 
-class FakeTime(object):
+class BaseFakeTime(object):
+    call_stack_inspection_limit = 5
 
-    def __init__(self, time_to_freeze, previous_time_function):
-        self.time_to_freeze = time_to_freeze
+    def _should_use_real_time(self):
+        if not self.call_stack_inspection_limit:
+            return False
+
+        if not self.ignore:
+            return False
+
+        frame = inspect.currentframe().f_back.f_back
+
+        for _ in range(self.call_stack_inspection_limit):
+            module_name = frame.f_globals.get('__name__')
+            if module_name and module_name.startswith(self.ignore):
+                return True
+
+            frame = frame.f_back
+            if frame is None:
+                break
+
+        return False
+
+
+class FakeTime(BaseFakeTime):
+
+    def __init__(self, previous_time_function, ignore=None):
         self.previous_time_function = previous_time_function
+        self.ignore = ignore
 
     def __call__(self):
-        current_time = self.time_to_freeze()
+        if self._should_use_real_time():
+            return real_time()
+        current_time = times_to_freeze[-1]()
         return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
 
 
-class FakeLocalTime(object):
-    def __init__(self, time_to_freeze, previous_localtime_function=None):
-        self.time_to_freeze = time_to_freeze
+class FakeLocalTime(BaseFakeTime):
+    def __init__(self, previous_localtime_function=None, ignore=None):
         self.previous_localtime_function = previous_localtime_function
+        self.ignore = ignore
 
     def __call__(self, t=None):
         if t is not None:
             return real_localtime(t)
-        shifted_time = self.time_to_freeze() - datetime.timedelta(seconds=time.timezone)
+        if self._should_use_real_time():
+            return real_localtime()
+        shifted_time = times_to_freeze[-1]() - datetime.timedelta(seconds=time.timezone)
         return shifted_time.timetuple()
 
 
-class FakeGMTTime(object):
-    def __init__(self, time_to_freeze, previous_gmtime_function):
-        self.time_to_freeze = time_to_freeze
+class FakeGMTTime(BaseFakeTime):
+    def __init__(self, previous_gmtime_function, ignore=None):
         self.previous_gmtime_function = previous_gmtime_function
+        self.ignore = ignore
 
     def __call__(self, t=None):
         if t is not None:
             return real_gmtime(t)
-        return self.time_to_freeze().timetuple()
+        if self._should_use_real_time():
+            return real_gmtime()
+        return times_to_freeze[-1]().timetuple()
 
 
-class FakeStrfTime(object):
-    def __init__(self, time_to_freeze, previous_strftime_function):
-        self.time_to_freeze = time_to_freeze
+class FakeStrfTime(BaseFakeTime):
+    def __init__(self, previous_strftime_function, ignore=None):
         self.previous_strftime_function = previous_strftime_function
+        self.ignore = ignore
 
     def __call__(self, format, time_to_format=None):
         if time_to_format is None:
-            time_to_format = FakeLocalTime(self.time_to_freeze)()
+            if not self._should_use_real_time():
+                time_to_format = FakeLocalTime()()
+
         return real_strftime(format, time_to_format)
 
 
+class FakeClock(BaseFakeTime):
+    def __init__(self, previous_clock_function, tick=False, ignore=None):
+        self.previous_clock_function = previous_clock_function
+        self.tick = tick
+        self.ignore = ignore
+
+    def __call__(self, *args, **kwargs):
+        if self._should_use_real_time():
+            return real_clock()
+
+        if len(times_to_freeze) == 1:
+            return 0.0 if not self.tick else self.previous_clock_function()
+
+        first_frozen_time = times_to_freeze[0]()
+        last_frozen_time = times_to_freeze[-1]()
+
+        timedelta = (last_frozen_time - first_frozen_time)
+        total_seconds = timedelta.total_seconds()
+
+        if self.tick:
+            total_seconds += self.previous_clock_function()
+
+        return total_seconds
+
+
 class FakeDateMeta(type):
     @classmethod
     def __instancecheck__(self, obj):
@@ -194,9 +252,6 @@ def date_to_fakedate(date):
 
 
 class FakeDate(with_metaclass(FakeDateMeta, real_date)):
-    dates_to_freeze = []
-    tz_offsets = []
-
     def __new__(cls, *args, **kwargs):
         return real_date.__new__(cls, *args, **kwargs)
 
@@ -220,13 +275,13 @@ class FakeDate(with_metaclass(FakeDateMeta, real_date)):
         result = cls._date_to_freeze() + cls._tz_offset()
         return date_to_fakedate(result)
 
-    @classmethod
-    def _date_to_freeze(cls):
-        return cls.dates_to_freeze[-1]()
+    @staticmethod
+    def _date_to_freeze():
+        return times_to_freeze[-1]()
 
     @classmethod
     def _tz_offset(cls):
-        return cls.tz_offsets[-1]
+        return tz_offsets[-1]
 
 FakeDate.min = date_to_fakedate(real_date.min)
 FakeDate.max = date_to_fakedate(real_date.max)
@@ -239,9 +294,6 @@ class FakeDatetimeMeta(FakeDateMeta):
 
 
 class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
-    times_to_freeze = []
-    tz_offsets = []
-
     def __new__(cls, *args, **kwargs):
         return real_datetime.__new__(cls, *args, **kwargs)
 
@@ -276,6 +328,13 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     def date(self):
         return date_to_fakedate(self)
+    
+    @property
+    def nanosecond(self):
+        try:
+            return real_datetime.nanosecond
+        except AttributeError:
+            return 0
 
     @classmethod
     def today(cls):
@@ -286,14 +345,15 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
         result = cls._time_to_freeze() or real_datetime.utcnow()
         return datetime_to_fakedatetime(result)
 
-    @classmethod
-    def _time_to_freeze(cls):
-        if cls.times_to_freeze:
-            return cls.times_to_freeze[-1]()
+    @staticmethod
+    def _time_to_freeze():
+        if times_to_freeze:
+            return times_to_freeze[-1]()
 
     @classmethod
     def _tz_offset(cls):
-        return cls.tz_offsets[-1]
+        return tz_offsets[-1]
+
 
 FakeDatetime.min = datetime_to_fakedatetime(real_datetime.min)
 FakeDatetime.max = datetime_to_fakedatetime(real_datetime.max)
@@ -343,6 +403,8 @@ def _parse_time_to_freeze(time_to_freeze_str):
         time_to_freeze = time_to_freeze_str
     elif isinstance(time_to_freeze_str, datetime.date):
         time_to_freeze = datetime.datetime.combine(time_to_freeze_str, datetime.time())
+    elif isinstance(time_to_freeze_str, datetime.timedelta):
+        time_to_freeze = datetime.datetime.utcnow() + time_to_freeze_str
     else:
         time_to_freeze = parser.parse(time_to_freeze_str)
 
@@ -466,23 +528,29 @@ class _freeze_time(object):
             time_to_freeze = FrozenDateTimeFactory(self.time_to_freeze)
 
         # Change the modules
+        is_already_started = len(times_to_freeze) > 0
+        times_to_freeze.append(time_to_freeze)
+        tz_offsets.append(self.tz_offset)
+
+        if is_already_started:
+            return time_to_freeze
+
         datetime.datetime = FakeDatetime
-        datetime.datetime.times_to_freeze.append(time_to_freeze)
-        datetime.datetime.tz_offsets.append(self.tz_offset)
-
         datetime.date = FakeDate
-        datetime.date.dates_to_freeze.append(time_to_freeze)
-        datetime.date.tz_offsets.append(self.tz_offset)
 
-        fake_time = FakeTime(time_to_freeze, time.time)
-        fake_localtime = FakeLocalTime(time_to_freeze, time.localtime)
-        fake_gmtime = FakeGMTTime(time_to_freeze, time.gmtime)
-        fake_strftime = FakeStrfTime(time_to_freeze, time.strftime)
+        fake_time = FakeTime(time.time, ignore=self.ignore)
+        fake_localtime = FakeLocalTime(time.localtime, ignore=self.ignore)
+        fake_gmtime = FakeGMTTime(time.gmtime, ignore=self.ignore)
+        fake_strftime = FakeStrfTime(time.strftime, ignore=self.ignore)
+        fake_clock = FakeClock(time.clock, tick=self.tick, ignore=self.ignore)
+
         time.time = fake_time
         time.localtime = fake_localtime
         time.gmtime = fake_gmtime
         time.strftime = fake_strftime
-        uuid._uuid_generate_time = None
+        time.clock = fake_clock
+        if uuid_generate_time_attr:
+            setattr(uuid, uuid_generate_time_attr, None)
         uuid._UuidCreate = None
         uuid._last_timestamp = None
 
@@ -497,6 +565,7 @@ class _freeze_time(object):
             ('real_localtime', real_localtime, 'FakeLocalTime', fake_localtime),
             ('real_strftime', real_strftime, 'FakeStrfTime', fake_strftime),
             ('real_time', real_time, 'FakeTime', fake_time),
+            ('real_clock', real_clock, 'FakeClock', fake_clock),
         ]
         self.fake_names = tuple(fake_name for real_name, real, fake_name, fake in to_patch)
         self.reals = dict((id(fake), real) for real_name, real, fake_name, fake in to_patch)
@@ -517,7 +586,7 @@ class _freeze_time(object):
                 elif (not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time')):
                     continue
 
-                module_attrs = _get_cached_module_attributes(mod_name, module)
+                module_attrs = _get_cached_module_attributes(module)
                 for attribute_name, attribute_value in module_attrs:
                     fake = fakes.get(id(attribute_value))
                     if fake:
@@ -527,12 +596,10 @@ class _freeze_time(object):
         return time_to_freeze
 
     def stop(self):
-        datetime.datetime.times_to_freeze.pop()
-        datetime.datetime.tz_offsets.pop()
-        datetime.date.dates_to_freeze.pop()
-        datetime.date.tz_offsets.pop()
+        times_to_freeze.pop()
+        tz_offsets.pop()
 
-        if not datetime.datetime.times_to_freeze:
+        if not times_to_freeze:
             datetime.datetime = real_datetime
             datetime.date = real_date
             copyreg.dispatch_table.pop(real_datetime)
@@ -568,14 +635,16 @@ class _freeze_time(object):
                         if real:
                             setattr(module, module_attribute, real)
 
-        time.time = time.time.previous_time_function
-        time.gmtime = time.gmtime.previous_gmtime_function
-        time.localtime = time.localtime.previous_localtime_function
-        time.strftime = time.strftime.previous_strftime_function
+            time.time = time.time.previous_time_function
+            time.gmtime = time.gmtime.previous_gmtime_function
+            time.localtime = time.localtime.previous_localtime_function
+            time.strftime = time.strftime.previous_strftime_function
+            time.clock = time.clock.previous_clock_function
 
-        uuid._uuid_generate_time = real_uuid_generate_time
-        uuid._UuidCreate = real_uuid_create
-        uuid._last_timestamp = None
+            if uuid_generate_time_attr:
+                setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
+            uuid._UuidCreate = real_uuid_create
+            uuid._last_timestamp = None
 
     def decorate_coroutine(self, coroutine):
         return wrap_coroutine(self, coroutine)
@@ -604,10 +673,16 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
     except NameError:
         string_type = str
 
-    if not isinstance(time_to_freeze, (type(None), string_type, datetime.date,
-        types.FunctionType, types.GeneratorType)):
+    acceptable_times = (type(None), string_type, datetime.date, datetime.timedelta,
+             types.FunctionType, types.GeneratorType)
+
+    if MayaDT is not None:
+        acceptable_times += MayaDT,
+
+    if not isinstance(time_to_freeze, acceptable_times):
         raise TypeError(('freeze_time() expected None, a string, date instance, datetime '
-                         'instance, function or a generator, but got type {0}.').format(type(time_to_freeze)))
+                         'instance, MayaDT, timedelta instance, function or a generator, but got '
+                         'type {0}.').format(type(time_to_freeze)))
     if tick and not _is_cpython:
         raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')
 
@@ -617,8 +692,14 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
     if isinstance(time_to_freeze, types.GeneratorType):
         return freeze_time(next(time_to_freeze), tz_offset, ignore, tick)
 
+    if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
+        return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
+                           tick, as_arg)
+
     if ignore is None:
         ignore = []
+    ignore = ignore[:]
+    ignore.append('nose.plugins')
     ignore.append('six.moves')
     ignore.append('django.utils.six.moves')
     ignore.append('google.gax')
